<?php
/**
 * Created by PhpStorm.
 * User: Pascal Roessler
 * Date: 05/22/23
 * Time: 10:45 AM
 */

namespace IMATHUZH\Qfq\Core\Form;

use CodeException;
use DbException;
use IMATHUZH\Qfq\Core\Form\FormElement\AbstractFormElement;
use IMATHUZH\Qfq\Core\Helper\HelperFormElement;
use IMATHUZH\Qfq\Core\Helper\KeyValueStringParser;
use IMATHUZH\Qfq\Core\Helper\OnArray;
use IMATHUZH\Qfq\Core\Helper\Path;
use IMATHUZH\Qfq\Core\Helper\Sanitize;
use UserFormException;
use UserReportException;

/**
 * Class FormElementInput extends AbstractFormElement
 * @package qfq
 */
class FormElementInput extends AbstractFormElementBackup {
    /**
     * @var array
     */
    protected array $colsRows = array();
    /**
     * @var string|null
     */
    protected ?string $step = null;
    /**
     * @var string|null
     */
    protected ?string $inputAutocomplete = null;
    /**
     * @var string|null
     */
    protected ?string $autofocus = null;
    /**
     * @var string|null
     */
    protected ?string $placeholder = null;
    /**
     * @var string|null
     */
    protected ?string $htmlFormElementName = null;
    /**
     * @var string|null
     */
    protected ?string $autocomplete = null;
    /**
     * @var string|null
     */
    protected ?string $clearMe = null;
    /**
     * @var string|null
     */
    public ?string $characterCountWrap = null;
    /**
     * @var string|null
     */
    public ?string $attributeCC = null;
    /**
     * @var string|null
     */
    public ?string $classCC = null;
    /**
     * @var string|null
     */
    protected ?string $inputType = null;
    /**
     * @var bool|null
     */
    public ?bool $flagTextarea = null;
    /**
     * @var string|null
     */
    public ?string $htmlTag = null;
    /**
     * @var string|null
     */
    public ?string $textAreaValue = null;
    /**
     * @var string|null
     */
    protected ?string $typeAheadSqlPrefetch = null;
    /**
     * @var string|null
     */
    protected ?string $typeAheadTag = null;
    /**
     * @var string|null
     */
    protected ?string $typeAheadInitialSuggestion = null;
    /**
     * @var string|null
     */
    private ?string $typeAheadTagDelimiter = null;
    /**
     * @var int
     */
    protected int $typeAheadLimit = TYPEAHEAD_DEFAULT_LIMIT;
    /**
     * @var int
     */
    protected int $typeAheadMinLength = 2;
    /**
     * @var string
     */
    protected string $ldapServer = '';
    /**
     * @var string
     */
    protected string $ldapBaseDn = '';
    /**
     * @var string
     */
    protected string $typeAheadLdapSearch = '';
    /**
     * @var string
     */
    protected string $typeAheadLdapSearchPrefetch = '';
    /**
     * @var string
     */
    protected string $typeAheadLdapValuePrintf = '';
    /**
     * @var string
     */
    protected string $typeAheadLdapIdPrintf = '';
    /**
     * @var string
     */
    protected string $ldapUseBindCredentials = '';
    /**
     * @var string
     */
    protected string $decimalFormat = '';
    /**
     * @var string|null
     */
    protected ?string $hideZero = null;
    /**
     * @var string|null
     */
    protected ?string $_retypeSourceName = null;
    /**
     * @var string|null
     */
    protected ?string $textareaResize = null;

    /**
     * @param mixed $formElement The form element data.
     * @param mixed $value The value for the form element.
     * @param string $htmlFormElementName The name attribute for the HTML form element.
     * @param mixed $dbIndex The database index data.
     *
     * @throws CodeException
     * @throws UserFormException
     * @throws DbException
     * @throws UserReportException
     */
    public function __construct($formElement, $value, $htmlFormElementName, $dbIndex) {
        parent::__construct($formElement, $dbIndex);

        $this->value = $value;
        $this->htmlFormElementName = $htmlFormElementName;
        $this->htmlClass[] = 'form-control';
        $this->dbIndexData = $dbIndex;
    }

    /**
     * This function performs various checks and manipulations on the formElement object.
     * Then fills the htmlAttributes array with all the necessary keys and values defined in the formEditor.
     *
     * @return void
     *
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     * @throws UserReportException
     */
    function process() {
        if (!empty($this->clearMe) && $this->clearMe != '0') {
            if (($this->formSpec[F_MULTI_MODE] ?? '') == 'vertical') {
                $this->htmlClass[] = 'qfq-clear-me-multiform';
            } else {
                $this->htmlClass[] = 'qfq-clear-me';
            }
        }

        if ($this->maxLength > 0 && $this->value !== '' && empty($this->typeAheadSql)) {
            // Check if there are \r\n > those should be counted as one char, not two. #13030
            $crlf = substr_count($this->value, "\r\n");
            $max = $this->maxLength + $crlf;
            // crop string only if it's not empty (substr returns false on empty strings)
            $this->value = mb_substr($this->value, 0, $max);
        }

        if ($this->hideZero != '0' && $this->value == '0') {
            $this->value = '';
        }

        if ($this->decimalFormat !== '') {
            if ($this->value !== '') { // empty string causes exception in number_format()
                $decimalScale = explode(',', $this->decimalFormat)[1]; // scale: Nachkommastellen
                $this->value = number_format($this->value, $decimalScale, '.', '');
            }
        }
        // Check if typeAhead[Tag] needs to build
        if ('' != ($typeAheadUrlParam = $this->typeAheadBuildParam())) {

            if (empty($this->inputType)) {
                //$formElement[FE_INPUT_TYPE] = FE_TYPE_SEARCH; // typeahead behaves better with 'search' instead of 'text'
            }

            if (empty($this->autocomplete)) {
                $this->autocomplete = 'off'; // typeahead behaves better with 'autocomplete'='off'
            }

            // Collect typeAhead initial suggestion
            if (!empty($this->typeAheadInitialSuggestion)) {
                $this->typeAheadMinLength = 0; // If a suggestion is defined: minLength becomes automatically 0
                $arr = $this->evaluate->parse($this->typeAheadInitialSuggestion);
                $arr = $this->dbArray[$this->dbIndexData]->makeArrayDict($arr, TYPEAHEAD_SQL_KEY_NAME, API_TYPEAHEAD_VALUE, API_TYPEAHEAD_KEY, API_TYPEAHEAD_VALUE);
                $this->htmlAttributes[DATA_TYPEAHEAD_INITIAL_SUGGESTION] = json_encode($arr);
            }

            $this->htmlClass[] = CLASS_TYPEAHEAD;
            $dataSip = $this->sip->queryStringToSip($typeAheadUrlParam, RETURN_SIP);

            $this->htmlAttributes[DATA_TYPEAHEAD_SIP] = $dataSip;
            $this->htmlAttributes[DATA_TYPEAHEAD_LIMIT] = $this->typeAheadLimit;
            $this->htmlAttributes[DATA_TYPEAHEAD_MINLENGTH] = $this->typeAheadMinLength;

            if (HelperFormElement::booleParameter($this->typeAheadPedantic ?? '-')) {
                $this->htmlAttributes[DATA_TYPEAHEAD_PEDANTIC] = 'true';
            }

            // Tag
            if (HelperFormElement::booleParameter($this->typeAheadTag ?? '-')) {
                $this->htmlAttributes[DATA_TYPEAHEAD_TAG] = 'true';

                //Default: tab, comma
                $kk = '[' . ($this->typeAheadTagDelimiter ?? '9,44') . ']';
                $this->htmlAttributes[DATA_TYPEAHEAD_TAG_DELIMITER] = $kk;
                $this->inputType = 'hidden';

                // Client: TAG handling expects the '$value' as a JSON string.
                $arr = KeyValueStringParser::parse($this->value, PARAM_KEY_VALUE_DELIMITER, PARAM_LIST_DELIMITER, KVP_IF_VALUE_EMPTY_COPY_KEY);
                $this->value = OnArray::arrayToQfqJson($arr);
            }
            // Give typeAhead url for ability to initialize dynamic html elements
            $this->htmlAttributes[DATA_TYPEAHEAD_URL] = Path::urlApi(API_TYPEAHEAD_PHP);
        }

        if (!empty($this->characterCountWrap)) {
            $this->htmlClass[] = CLASS_CHARACTER_COUNT;
            $this->htmlAttributes[DATA_CHARACTER_COUNT_ID] = $this->htmlId . HTML_ID_EXTENSION_CHARACTER_COUNT;
            $this->attributeCC = $this->htmlId . HTML_ID_EXTENSION_CHARACTER_COUNT;
            $this->classCC = ($this->characterCountWrap == '') ? 'qfq-cc-style' : '';

        }

        if (!empty($this->_retypeSourceName)) {
            $htmlFormElementNamePrimary = str_replace(FE_RETYPE_NAME_EXTENSION, '', $this->htmlFormElementName);
            $this->htmlAttributes[FE_DATA_MATCH] = '[name=' . str_replace(':', '\\:', $htmlFormElementNamePrimary) . ']';
        }

        // 'maxLength' needs an upper 'L': naming convention for DB tables!
        if ($this->maxLength > 0) {
            $this->htmlAttributes[FE_MAX_LENGTH] = [
                FE_VALUE => $this->maxLength,
                FLAG_OMIT_EMPTY => false
            ];
        }

        // In case the user specifies MIN & MAX with numbers, the html tag 'type' has to be 'number', to make the range check work in the browser.
        if (empty($this->inputType) && !empty($this->min) && !empty($this->max) && is_numeric($this->min) && is_numeric($this->max)) {
            $this->inputType = 'number';
        }

        // Check for input type 'textarea'. Possible notation: a) '', b) '<width>', c) '<width>,<height>', d) '<width>,<min-height>,<max-height>'
        $this->colsRows = explode(',', $this->size, 3);

        if (empty($this->colsRows[1])) {
            $this->colsRows[1] = 1;
        }

        // Active Textarea if min-height>1 (rows) or third parameter max-height>0
        $this->flagTextarea = ($this->colsRows[1] > 1 || ($this->colsRows[2] ?? '') > 0);

        $this->renderer->prepareExtraButton($this, !$this->flagTextarea);

        if ($this->flagTextarea) {
            $this->htmlTag = 'textarea';

            $this->htmlAttributes[ATTRIBUTE_COLS] = $this->colsRows[0];
            $this->htmlAttributes[ATTRIBUTE_ROWS] = $this->colsRows[1];

            if (($this->textareaResize ?? '1') == '0') {
                $this->htmlAttributes[ATTRIBUTE_STYLE] = 'resize: none;';
            }
            // Check if $colsRows[2] != 0 >> enable auto-grow (if nothing is defined: enable auto-grow)
            $maxHeight = $colsRows[2] ?? '350';
            if ($maxHeight != '0') {
                $this->htmlClass[] = CLASS_FORM_ELEMENT_AUTO_GROW;
                $this->htmlAttributes[ATTRIBUTE_DATA_MAX_HEIGHT] = $maxHeight;
            }

            $this->textAreaValue = htmlentities($this->value);

        } else {
            $this->htmlTag = FE_INPUT;
            if (!empty($this->inputType)) {
                $this->type = $this->inputType;

                // TypeAhead tag elements needs to be hidden
                //if (HelperFormElement::booleParameter($formElement[FE_TYPEAHEAD_TAG] ?? '-')) {
                //    $formElement[FE_TYPE] = 'hidden';
                //}
            }
            // If type password is selected then type text with own class will be taken to fake password over CSS
            if ($this->type === 'password') {
                $this->type = 'text';
                $this->htmlClass[] = 'qfq-password';
            }

            $this->htmlAttributes[FE_TYPE] = $this->type;
            $this->htmlAttributes[FE_SIZE] = $this->size;

            // Typeahead prefetch needs key as attribute. Key/value pair is needed later on client side.
            if (!empty($this->typeAheadSqlPrefetch) && is_array($this->value)) {
                $attributeValue = OnArray::getFirstValueFromArray($this->value, true);
            } else {
                $attributeValue = $this->value;
            }

            $this->htmlAttributes[FE_VALUE] = [
                FE_VALUE => htmlentities($attributeValue),
                FLAG_OMIT_EMPTY => false
            ];

        }

        $this->checkPattern = Sanitize::getInputCheckPattern($this->checkType, $this->checkPattern, $this->decimalFormat, $sanitizeMessage);
        if ($this->checkPattern !== '') {
            $this->htmlAttributes[ATTRIBUTE_PATTERN] = [
                FE_VALUE => $this->checkPattern,
                FLAG_OMIT_EMPTY => false,
                MODE_ESCAPE => ESCAPE_WITH_BACKSLASH
            ];
            // Use system specific message only,if no custom one is set.
            if ($this->data_pattern_error == $this->data_pattern_error_system) {
                $this->data_pattern_error = $sanitizeMessage;
            }
            $this->htmlAttributes[F_FE_DATA_PATTERN_ERROR] = [
                FE_VALUE => $this->data_pattern_error,
                FLAG_OMIT_EMPTY => false,
                MODE_ESCAPE => ESCAPE_WITH_BACKSLASH
            ];
        }

        if ($this->mode == FE_MODE_REQUIRED) {
            $this->htmlAttributes[F_FE_DATA_REQUIRED_ERROR] = $this->data_required_error;
        }

        if (empty($this->data_match_error)) {
            $this->data_required_error = F_FE_DATA_REQUIRED_ERROR_DEFAULT;
        }

        $this->getAttributeFeMode($this->mode, $this->htmlAttributes, false);
        $this->htmlAttributes[FE_CLASS] = implode(" ", $this->htmlClass);
        $this->htmlAttributes[FE_ID] = $this->htmlId;
        $this->htmlAttributes[FE_NAME] = $this->htmlFormElementName;
        $this->htmlAttributes[ATTRIBUTE_DATA_REFERENCE] = $this->dataReference;
        $this->htmlAttributes[FE_INPUT_AUTOCOMPLETE] = $this->inputAutocomplete;
        $this->htmlAttributes[FE_AUTOFOCUS] = $this->autofocus;
        $this->htmlAttributes[FE_PLACEHOLDER] = $this->placeholder;
        $this->htmlAttributes[F_FE_DATA_MATCH_ERROR] = $this->data_match_error;
        $this->htmlAttributes[F_FE_DATA_ERROR] = $this->data_error;
        $this->htmlAttributes[FE_MIN] = $this->min;
        $this->htmlAttributes[FE_MAX] = $this->max;
        $this->htmlAttributes[FE_STEP] = $this->step;
        $this->htmlAttributes[FE_DATA_LOAD] = ($this->dynamicUpdate === 'yes') ? FE_DATA_LOAD : '';
        $this->htmlAttributes[F_TITLE] = $this->tooltip;
        //$this->htmlAttributes[FE_MODE] = $this->mode;

        // id and and name should be at the start of the array
        $order = [FE_ID, FE_NAME];

        $newAttributes = [];

        // Add id and id to the new array
        if (isset($this->htmlAttributes[FE_ID])) {
            $newAttributes[FE_ID] = $this->htmlAttributes[FE_ID];
        }
        if (isset($this->htmlAttributes[FE_NAME])) {
            $newAttributes[FE_NAME] = $this->htmlAttributes[FE_NAME];
        }

        // Add the remaining elements to the new array
        foreach ($this->htmlAttributes as $key => $value) {
            if (!in_array($key, $order)) {
                $newAttributes[$key] = $value;
            }
        }

        // Assign the correctly ordered array back to $this->htmlAttributes
        $this->htmlAttributes = $newAttributes;

    }

    /**
     * Check $formElement for FE_TYPE_AHEAD_SQL or FE_TYPE_AHEAD_LDAP_SERVER.
     * If one of them is given: build $urlParam with typeAhead Params.
     * Additionally set some parameter for later outside use, especially FE_TYPEAHEAD_LIMIT, FE_TYPEAHEAD_MINLENGTH
     *
     * @return string
     *
     * @throws CodeException
     * @throws UserFormException
     * @throws DbException
     * @throws UserReportException
     */
    private function typeAheadBuildParam(): string {

        $arr = [];

        if (!empty($this->typeAheadSql)) {
            $sql = $this->checkSqlAppendLimit($this->typeAheadSql, $this->typeAheadLimit);

            $arr = [
                FE_TYPEAHEAD_SQL => $sql,
                FE_TYPEAHEAD_SQL_PREFETCH => $this->typeAheadSqlPrefetch ?? ''
            ];
        } elseif (!empty($this->typeAheadLdap)) {
            foreach ([FE_LDAP_SERVER, FE_LDAP_BASE_DN, FE_TYPEAHEAD_LDAP_SEARCH] as $key) {
                // PR: unsure if $key works was $formElement[$key] before
                if ($this->{$key} == '') {
                    throw new UserFormException('Missing definition: ' . $key, ERROR_MISSING_DEFINITON);
                }
            }

            if ($this->typeAheadLdapValuePrintf . $this->typeAheadLdapIdPrintf == '') {
                throw new UserFormException('Missing definition: ' . FE_TYPEAHEAD_LDAP_VALUE_PRINTF . ' or ' . FE_TYPEAHEAD_LDAP_ID_PRINTF, ERROR_MISSING_DEFINITON);
            }

            $arr = [
                FE_LDAP_SERVER => $this->ldapServer,
                FE_LDAP_BASE_DN => $this->ldapBaseDn,
                FE_TYPEAHEAD_LDAP_SEARCH => $this->typeAheadLdapSearch,
                FE_TYPEAHEAD_LDAP_SEARCH_PREFETCH => $this->typeAheadLdapSearchPrefetch,
                FE_TYPEAHEAD_LDAP_VALUE_PRINTF => $this->typeAheadLdapValuePrintf,
                FE_TYPEAHEAD_LDAP_ID_PRINTF => $this->typeAheadLdapIdPrintf,
                FE_TYPEAHEAD_LIMIT => $this->typeAheadLimit,
            ];

            if (!empty($this->typeAheadLdapSearchPerToken)) {
                $arr[FE_TYPEAHEAD_LDAP_SEARCH_PER_TOKEN] = '1';
            }

            if ($this->ldapUseBindCredentials == '1') {
                $arr[SYSTEM_LDAP_1_RDN] = $this->store->getVar(SYSTEM_LDAP_1_RDN, STORE_SYSTEM);
                $arr[SYSTEM_LDAP_1_PASSWORD] = $this->store->getVar(SYSTEM_LDAP_1_PASSWORD, STORE_SYSTEM);
            }

        } else {
            // Add code here to handle the case when neither type-ahead SQL nor type-ahead LDAP is specified.
            // This block will be executed when no type-ahead functionality is configured for the form element.
        }

        return OnArray::toString($arr);
    }

    /**
     * Checks if $sql contains a SELECT statement.
     * Check for existence of a LIMIT Parameter. If not found add one.
     *
     * @param string $sql
     * @param int $limit
     *
     * @return string   Checked and maybe extended $sql statement.
     *
     * @throws UserFormException
     */
    private function checkSqlAppendLimit($sql, $limit): string {
        $sqlTest = '';

        $sql = trim($sql);
        // If exist, unwrap  '{{!', '}}'
        if (substr($sql, 0, 2) == '{{') {
            $sql = trim(substr($sql, 2, strlen($sql) - 4));

            if ($sql[0] ?? '' == '!') {
                $sql = trim(substr($sql, 1));
            }
        }

        if ($sql[0] == '[') {
            // Remove optional existing dbIndex token.
            $pos = strpos($sql, ']');
            $sqlTest = substr($sql, $pos + 1);
        } else {
            $sqlTest = $sql;
        }

        if (false === stristr(substr($sqlTest, 0, 7), 'SELECT ')) {
            throw new UserFormException(
                json_encode([ERROR_MESSAGE_TO_USER => '"Expect a SELECT statement', ERROR_MESSAGE_TO_DEVELOPER => "Expect a SELECT statement in " . FE_TYPEAHEAD_SQL . " - got: " . $sqlTest]),
                ERROR_BROKEN_PARAMETER);

        }

        if (false === stristr($sql, ' LIMIT ')) {
            $sql .= " LIMIT $limit";
        }

        return $sql;
    }
}